package Dns;

import java.net.*;
import java.io.*;
import Dns.Msg.*;
import Msg.PacketMsg;
import java.util.*;

/* La classe resolver si occupa delle 
 richieste da parte di processi locali (non accetta richieste 
 esterne) e le invia al Name Server del dominio in cui si trova.
 In questo modo  possibile gestire una cache unica per ogni
 macchina.
 Il resolver si mette in ascolto sulla porta 5656.
 Fa partire un Thread per ogni richiesta.	
*/

public class Resolver {
	protected static final String LogFile="dnshosts.log";
	protected static final int TimeOutSec=20;
	protected InetAddress local;
	//timestamp unico per tutti i thread: assicura l'unicit dei messaggi
	protected int timestamp=0;
	//per prove locali:
	public int port;
	//protected static final int port=5656;
	protected Cache cache=new Cache();
	Value[] nameserver;
	int select=0;
	
	//costrutttore che specifica gli indirizzi dei Name Server
	public Resolver(Value[] nameserver) throws Exception{
		local=InetAddress.getLocalHost();
		this.nameserver=nameserver;	
	}//costruttore
	
	public Resolver(int port) throws Exception{
		this(LogFile,port);
	}
	
	public Resolver(String nsfile) throws Exception{
		this(nsfile,NameServer.DnsPort);
	}
	
	public Resolver() throws Exception{
		this(LogFile,NameServer.DnsPort);
	}
	
	//costruttore che legge da file (alla prima esecuzione lo crea)
	public Resolver(String nsfile,int port) throws Exception{
		File file=new File(nsfile);
		this.port=port;
		Value[] ns;
		if(file.exists()){
			//legge gli indirizzi dns dal file
			ObjectInputStream ois=new ObjectInputStream(new FileInputStream(file));
			ns=(Value[])ois.readObject();
		}else{
			//prima esecuzione
			System.out.println("Inserire \"host:port\" per i Dns del dominio (invio per finire)");
			String inputln="";
			StringTokenizer st;
			ArrayList list=new ArrayList(); 
			do{
				BufferedReader stdIn = new BufferedReader(new InputStreamReader(System.in));
        inputln = stdIn.readLine();
				if(!(inputln.equals(""))){
					st=new StringTokenizer(inputln," :",false);
					String host=st.nextToken();
					int prport=0;
					try{
						prport=Integer.parseInt(st.nextToken());
					} catch(NumberFormatException nfe){
						System.out.println("errore "+ nfe);
						continue;
					}
					list.add(new Value(InetAddress.getByName(host),prport));
					System.out.println("Aggiunto NS "+InetAddress.getByName(host)+":"+prport);	
				}
			}while(!(inputln.equals("")));
			Object[] obj=list.toArray();
			ns=new Value[obj.length];
			for(int i=0;i<ns.length;i++){
				ns[i]=(Value)obj[i];			
			}
			if (ns.length==0) throw new Exception("Almeno un indirizzo NS necessario");
			ObjectOutputStream oos=new ObjectOutputStream(new FileOutputStream(file));
			oos.writeObject(ns);
		}//if
		local=InetAddress.getLocalHost();
		this.nameserver=ns;
	}//costruttore
	
	
	//il resolver si mette in ascolto:
	public void start(){
		
		//porta di ascolto per richieste locali
		DatagramSocket listsock=null;
		try{
			listsock=new DatagramSocket(port);
			System.out.println("Resolver connesso su "+port);
		} catch (Exception e){System.err.println("Errore in apertura socket:"+e.getMessage());}
		while(true){
			try {
				DatagramPacket packet=new DatagramPacket(new byte[PacketMsg.dimPacket],PacketMsg.dimPacket);
				listsock.receive(packet);
				if(!(packet.getAddress().equals(local))){
					System.out.println("Pacchetto proveniente da "+packet.getAddress()+" rifiutato.");
					continue;
				}
				QueryMsg msg=(QueryMsg)PacketMsg.readFromPacket(packet);
				System.out.println("*****************************************************");
				System.out.println("ricevuto messaggio : "+msg);
				(new ResolverThread(msg,packet.getAddress(),packet.getPort(),this)).start();
			} catch (ClassCastException cce){
				System.err.println("Ricevuto Messaggio Sconosciuto:"+cce);
			} catch (StreamCorruptedException sce){
				System.err.println("Errore nella lettura del messaggio ricevuto:"+sce);
			}catch (Exception e){
				System.err.println("Errore in ricezione richiesta:"+e);
			} //try
		}//while
	}//run 

	//metodo che i thread devono richiamare per ottenere
	//un timestamp con cui identificare univocamente i messaggi:
	public synchronized int stamp(){
		if (++timestamp==Integer.MAX_VALUE)timestamp=0;
		return timestamp;
	}
	
	//Distribuisce uniformemente le richieste tra i DNS:
	public Value getNameServer(){
 		return nameserver[(select++)%(nameserver.length)];
	}
	
	public static void main(String args[]){
		try{
			if(args.length==2){
				Resolver res=new Resolver(args[0],Integer.parseInt(args[1]));
				res.start();
			} else System.out.println("Usage: java Dns.SecondaryServer LogFile port");
		}catch(Exception e){
			System.out.println("Errore nello startup del Resolver:"+e);
		}
	}
	
	public Value[] cacheLookup(DNSName query){
		//ricerco il nome richiesto
		return cache.get(query);
	}
	
	public void cache(Binding b,int TTL){
		//inserisco il nome richiesto
		cache.put(b,TTL);
	}
	
	public int getCacheTimeToLive(DNSName query){
		return cache.getTimeToLive(query);
	}
	
}//class Resolver

class ResolverThread extends Thread {
	QueryMsg query;
	int querytimestamp;
	InetAddress retaddr;
	int retport;
	Resolver parent;
	
	public ResolverThread(QueryMsg msg,InetAddress retaddr,int retport ,Resolver parent){
		query=msg;
		querytimestamp=msg.getTimestamp();
		this.retaddr=retaddr;
		this.retport=retport;
		this.parent=parent;
	}

	
	public void run(){
		int select=0;
		
		//Apertura socket invio e ricezione risposta		
		DatagramSocket socket=null;
		DatagramPacket spacket=null,rpacket=null;
		PacketMsg retmsg=null;
		boolean done=false;
		
		try{
			socket=new DatagramSocket();
			socket.setSoTimeout(parent.TimeOutSec*1000);
		} catch (SocketException e){
			System.err.println("Errore in apertura socket:"+e.getMessage());
			stop();
		}

			//ricerca nella cache:
			Value[] cachefound=null;
			if(query instanceof NameQueryMsg)
					parent.cacheLookup(((NameQueryMsg)query).getQuery());
			
			if(cachefound==null){
				int timestamp=parent.stamp();
				QueryMsg msg;
				if(query instanceof NameQueryMsg){
					msg=new NameQueryMsg(((NameQueryMsg)query).getQuery(),true,timestamp);
				} else {
					msg=new DomainQueryMsg(timestamp);
				}
				Value ns=null;
				while(!done){
					try{
						ns=parent.getNameServer();
						spacket=msg.writeToPacket(ns.getAddress(),ns.getPort());
						rpacket=new DatagramPacket(new byte[PacketMsg.dimPacket],PacketMsg.dimPacket);
						socket.send(spacket);
		  			socket.receive(rpacket);				
						retmsg=SignedMsg.readFromPacket(rpacket);
						System.out.println(retmsg);
						if(((TimestampedMsg)retmsg).getTimestamp()!=timestamp){
							System.out.println("Errore timestamp non corretto");
							break;
						} else	{
							//caching
							if(retmsg instanceof BindFoundMsg){
								if(((BindFoundMsg)retmsg).getName().equals(((NameQueryMsg)query).getQuery())){
									Value[] retval=((CachableMsg)retmsg).getValue();
									for(int i=0;i<retval.length;i++){
										parent.cache(new Binding(((NameQueryMsg)query).getQuery(),retval[i]),((CachableMsg)retmsg).getTimeToLive());
									}
								}else{
									System.out.println("Nome ricevuto non corretto");
									break;
								}
							}//if
							((TimestampedMsg)retmsg).setTimestamp(querytimestamp);
							done=true;
						}
					}catch(InterruptedIOException iioe){
						System.out.println("Timeout scaduto");
						done=false;
					}catch(Exception e){
						System.out.println("Errore "+e);
						done=false;
						break;
					}
				}//while
			} else {
				//risposta trovata in cache:
				retmsg=new BindFoundMsg(((NameQueryMsg)query).getQuery(),cachefound,parent.getCacheTimeToLive(((NameQueryMsg)query).getQuery()),querytimestamp);
				done=true;
			}
			if(done){
				try{
					spacket=retmsg.writeToPacket(retaddr,retport);
					socket.send(spacket);
				}catch(Exception e){
					System.out.println("Impossibile inviare risposta");
				}
			}//if done

	}

}
